@@ -21,6 +21,8 @@ import time
 import glob
 import csv
 
+
+from contextlib import suppress
 from wheel_filename import parse_wheel_filename
 
 from .utils import *
@@ -302,7 +304,8 @@ class TerrariumAssembler:
             'install-repos': 'install RPM repositories',
             'install-rpms': 'install downloaded RPMS',
             'download-wheels': 'download needed WHL-python packages',
-            'preinstall-wheels': 'Bootstrap build environment with external wheel packages',
+            # 'preinstall-wheels': 'Bootstrap build environment with external wheel packages',
+            'init-env': 'Create  build environment with some bootstrapping',
             'build-wheels': 'compile wheels for our python sources',
             'install-wheels': 'Install our and external Python wheels',
             'build-nuitka': 'Compile Python packages to executable',
@@ -334,7 +337,8 @@ class TerrariumAssembler:
 
         if self.args.stage_build_and_pack:
             self.args.stage_install_rpms = True
-            self.args.stage_preinstall_wheels = True
+            # self.args.stage_preinstall_wheels = True
+            self.args.stage_init_env = True
             self.args.stage_build_wheels = True
             self.args.stage_install_wheels = True
             self.args.stage_build_nuitka = True
@@ -344,7 +348,8 @@ class TerrariumAssembler:
         if self.args.stage_my_source_changed:
             # self.args.stage_checkout = True
             self.args.stage_download_wheels = True
-            self.args.stage_preinstall_wheels = True
+            self.args.stage_init_env = True
+            # self.args.stage_preinstall_wheels = True
             self.args.stage_build_wheels = True
             self.args.stage_install_wheels = True
             self.args.stage_build_nuitka = True
@@ -999,6 +1004,11 @@ popd
         os.remove(patched_binary)
         pass
 
+    def get_all_sources(self):
+        for td_ in self.projects() + self.spec.templates_dirs:
+            git_url, git_branch, path_to_dir_, _ = self.explode_pp_node(td_)
+            yield git_url, git_branch, path_to_dir_
+
     def folder_command(self):
         '''
             Just checking out sources.
@@ -1010,8 +1020,10 @@ popd
         curdir = os.getcwd()
         args = self.args
         in_src = os.path.relpath(self.src_dir, start=self.curdir)
-        for td_ in self.projects() + self.spec.templates_dirs:
-            git_url, git_branch, path_to_dir_, _ = self.explode_pp_node(td_)
+        # for td_ in self.projects() + self.spec.templates_dirs:
+        #     git_url, git_branch, path_to_dir_, _ = self.explode_pp_node(td_)
+
+        for git_url, git_branch, path_to_dir_ in self.get_all_sources():
             os.chdir(curdir)
             if os.path.exists(path_to_dir_):
                 os.chdir(path_to_dir_)
@@ -1084,7 +1096,7 @@ if [ -d "%(newpath)s" ]; then
 fi            
 """ % vars())
 
-        self.lines2sh("06-checkout", lines, 'checkout')    
+        self.lines2sh("05-checkout", lines, 'checkout')    
         # self.lines2sh("96-pullall", lines2)    
         pass
 
@@ -1143,7 +1155,14 @@ fi
         root_dir = self.root_dir
         os.chdir(self.curdir)
 
-        os.chdir(os.path.join('in', 'src', 'pip'))
+        pipdir = ''
+        for pdir in ('github-belonesox-pip', 'pip'):
+            pipdir = os.path.join('in', 'src', pdir)
+            if os.path.exists(pipdir):
+                break
+
+        os.chdir(pipdir)    
+        # os.chdir(os.path.join('in', 'src', 'github-belonesox-pip'))
         scmd = f'''{self.root_dir}/ebin/python3 setup.py install --single-version-externally-managed --root / '''
         os.system(scmd)
 
@@ -1341,23 +1360,40 @@ pipenv run sh -c "pushd {path_to_dir};python3 setup.py clean --all;python3 setup
 """
             lines.append(scmd)
             pass
-        self.lines2sh("09-build-wheels", lines, "build-wheels")
+        self.lines2sh("06-build-wheels", lines, "build-wheels")
 
 
-    def preinstall_wheels(self):
+#     def preinstall_wheels(self):
+#         os.chdir(self.curdir)
+
+#         ext_whl_path = os.path.relpath(self.ext_whl_path, self.curdir)
+
+#         lines = []
+#         scmd = f'''
+# pipenv --rm
+# rm -f Pipfile*
+# pipenv install --python {self.tvars.python_version_1}.{self.tvars.python_version_2}
+# pipenv run python3 -m pip install ./in/bin/extwheel/*.whl --find-links="{ext_whl_path}"  --force-reinstall --ignore-installed  --no-cache-dir --no-index 
+# ''' 
+#         lines.append(scmd)   
+#         self.lines2sh("08-preinstall-wheels", lines, "preinstall-wheels")
+#         pass    
+
+    def init_env(self):
         os.chdir(self.curdir)
 
         ext_whl_path = os.path.relpath(self.ext_whl_path, self.curdir)
 
         lines = []
         scmd = f'''
-pipenv --rm
+python -m pipenv --rm
 rm -f Pipfile*
-pipenv install --python {self.tvars.python_version_1}.{self.tvars.python_version_2}
-pipenv run python3 -m pip install ./in/bin/extwheel/*.whl --find-links="{ext_whl_path}"  --force-reinstall --ignore-installed  --no-cache-dir --no-index 
+# sudo python -m pip uninstall -y virtualenv
+python -m pipenv install --python {self.tvars.python_version_1}.{self.tvars.python_version_2}
+python -m pipenv install setuptools_git_versioning wheel
 ''' 
         lines.append(scmd)   
-        self.lines2sh("08-preinstall-wheels", lines, "preinstall-wheels")
+        self.lines2sh("04-init-env", lines, "init-env")
         pass    
 
     def install_wheels(self):
@@ -1461,7 +1497,7 @@ rm -f %s/extwheel/*
 
         pip_args_ = self.pip_args_from_sources()
 
-        scmd = f"python3 -m pip download wheel {pip_args_} --dest {bin_dir}/extwheel " 
+        scmd = f"python3 -m pip download wheel {pip_args_} --dest {bin_dir}/extwheel --find-links='{bin_dir}/ourwheel' "  
         lines.append(scmd)                
 
         scmd = f"rm -f {bin_dir}/extwheel/enum34* "
@@ -1471,6 +1507,8 @@ rm -f %s/extwheel/*
 pushd {bin_dir}/extwheel
 ls *.tar.* | xargs -i[] -t python3 -m pip wheel [] --no-deps
 rm -f *.tar.*
+popd
+python -c "import os; whls = [d.split('.')[0]+'*' for d in os.listdir('{bin_dir}/ourwheel')]; os.system('cd {bin_dir}/extwheel; rm -f ' + ' '.join(whls))"
 """ 
         lines.append(scmd)                
         self.lines2sh("07-download-wheels", lines, "download-wheels")
@@ -1699,7 +1737,13 @@ rm -f *.tar.*
                             linkto = os.readlink(fname_)
                             os.symlink(linkto, out_fname_)
                         else:    
-                            if plain or fname_.endswith('.nj2'):
+                            if fname_.endswith('.copy-file'):
+                                out_fname_ = os.path.splitext(out_fname_)[0]
+                                path_ = open(fname_).read().strip()
+                                if not os.path.isabs(path_):
+                                    path_ = os.path.join(self.curdir, path_)
+                                shutil.copy2(path_, out_fname_)
+                            elif plain or fname_.endswith('.nj2'):
                                 template = env.get_template(fname_)
                                 output = template.render(self.tvars)                    
                                 with open(out_fname_, 'w', encoding='utf-8') as lf_:
@@ -1742,7 +1786,8 @@ rm -f *.tar.*
         self.download_packages()
         self.checkout_sources()
         self.download_pip()
-        self.preinstall_wheels()
+        # self.preinstall_wheels()
+        self.init_env()
         self.build_wheels()
         self.install_wheels()
         self.build_nuitkas()
@@ -1991,12 +2036,48 @@ terrarium_assembler --debug --stage-pack=./out-debug "%(specfile_)s" --stage-mak
             print("Size ", size_/1024/1024, 'Mb')
 
         if self.args.stage_make_isoexe:
+            from dateutil.relativedelta import relativedelta            
+
             os.chdir(self.curdir)
-            time_prefix = datetime.datetime.now().replace(microsecond=0).isoformat().replace(':', '-')
+
+            isodir = root_dir + '.iso'
+            mkdir_p(isodir)
+            old_isos = [f_ for f_ in os.listdir(isodir) if f_.endswith('.iso')]
+            
+            current_time = datetime.datetime.now().replace(microsecond=0) 
+            prev_release_time = current_time + relativedelta(months=-1)
+            
+            for iso_ in reversed(sorted(old_isos)):
+                try:
+                    prev_release_time = datetime.datetime.strptime(iso_[:19], '%Y-%m-%dT%H-%M-%S')
+                    break
+                except:    
+                    pass
+
+            since_time_ = prev_release_time.isoformat()
+            gitlogcmd_ = f'git log --since="{since_time_}" --pretty --name-status '
+
+            lines_ = []
+            for git_url, git_branch, path_to_dir_ in self.get_all_sources():
+                os.chdir(self.curdir)
+                if os.path.exists(path_to_dir_):
+                    os.chdir(path_to_dir_)
+                    with suppress(Exception):
+                        change_ = subprocess.check_output(gitlogcmd_, shell=True).decode('utf-8').strip()
+                        if change_:
+                            lines_.append(f'----\n Changelog for {path_to_dir_} ({git_url} / {git_branch})')
+                            lines_.append(change_)
+
+            pass
+
+
+            # current_time = datetime.datetime.now().replace(microsecond=0) 
+            time_prefix = current_time.isoformat().replace(':', '-')
             label = 'disk'
             if 'label' in self.spec:
                 label = self.spec.label
             installscript = "install-me.sh" % vars()
+            os.chdir(self.curdir)
             installscriptpath = os.path.abspath(os.path.join("tmp/", installscript))
             if os.path.exists(installscriptpath):
                 os.unlink(installscriptpath)
@@ -2006,22 +2087,39 @@ terrarium_assembler --debug --stage-pack=./out-debug "%(specfile_)s" --stage-mak
                 pmode = ' --threads 8 --pbzip2 ' 
             os.chdir(self.curdir)
             self.cmd(f'chmod a+x {root_dir}/install-me')
+
+            filename = f"{time_prefix}-{label}-dm.iso" % vars()
+            with suppress(Exception):
+                chp_ = os.path.join(root_dir, 'isodistr.txt')
+                open(chp_, 'w', encoding='utf-8').write(filename)
+
             scmd = (f'''
             makeself.sh {pmode} --needroot {root_dir} {installscriptpath} "Installation" ./install-me             
         ''' % vars()).replace('\n', ' ').strip()
             if not self.cmd(scmd)==0:
                 print(f'« {scmd} » failed!')
                 return
+            os.chdir(self.curdir)
+            changelogfilename = filename + '.changelog.txt'
 
-            filename = "%(time_prefix)s-%(label)s-dm.iso" % vars()
-            isodir = root_dir + '.iso'
-            mkdir_p(isodir)
             filepath = os.path.join(isodir, filename)
             scmd = ('''
         mkisofs -r -J -o  %(filepath)s  %(installscriptpath)s 
         ''' % vars()).replace('\n', ' ').strip()
-            print(scmd)
-            os.system(scmd)
+            os.chdir(self.curdir)
+            self.cmd(scmd)
+            scmd = (f'''
+        md5sum {filepath}
+        ''').replace('\n', ' ').strip()
+            os.chdir(self.curdir)
+            md5s_ = subprocess.check_output(scmd, shell=True).decode('utf-8').strip().split()[0]
+            lines_.insert(0, f';MD5: {md5s_}')
+
+            with suppress(Exception):
+                chp_ = os.path.join(isodir, changelogfilename)
+                open(chp_, 'w', encoding='utf-8').write('\n'.join(lines_))
+                open(f'{filepath}.md5', 'w', encoding='utf-8').write(md5s_)
+
             os.chdir(isodir)
             scmd = f'''ln -sf {filename} last.iso'''
             self.cmd(scmd)
